!***************************************
! coarsepm.f90 - coarse mesh subroutines 
! Copyright (C) 2004 Hugh Merz - merz@cita.utoronto.ca
! latest revision Dec 14 2004
! -- added coarse density power spectrum
!***********************************************************
!This program is free software; you can redistribute it and/or
!modify it under the terms of the GNU General Public License
!as published by the Free Software Foundation; either version 2
!of the License, or (at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


  subroutine coarsepm
    implicit none
    
    include 'globalpm.fh'

#ifdef DEBUG_KERN
    call density_check
#endif

    call coarse_force 

if (master) write(*,*) 'finished coarse force:', t_elapsed(wcts)

    if (coarse_ps) then
      call coarse_power
      if (master) write(*,*) 'finished coarse power:', t_elapsed(wcts)
    endif    
   
    call force_buf

if (master) write(*,*) 'finished force buffer:', t_elapsed(wcts)

    call max_coarse_dt

if (master) write(*,*) 'finished coarse dt:', t_elapsed(wcts)

    call coarse_velocity

if (master) write(*,*) 'finished coarse velocity:', t_elapsed(wcts)

  end subroutine coarsepm

!*******
#ifdef DEBUG_KERN
  subroutine density_check
   
    implicit none
    include 'mpif.h'
    include 'globalpm.fh'

    integer(kind=4) :: i,j,k,ierr
    real(kind=8) :: partsum,partot

    !$omp parallel do default(shared) private(i,j,k) reduction(+:partsum)
    do k=1,LC/NN
      do j=1,LC
        do i=1,LC
          partsum=partsum+real(rhoc(i,j,k),kind=8)
        enddo
      enddo
    enddo
    !$omp end parallel do
    if (nts == 1) then 
    open(156,file='/mnt/s0/pmfast/fftw_test/rhoc_file.dat')
    do k=1,LC/NN
      do j=1,LC
        do i=1,LC
          write(156,'(f20.10)') rhoc(i,j,k)
        enddo
      enddo
    enddo
    close(156)
    endif
 
    partsum=partsum/real(LPS,kind=8)**3
    write(*,*) 'mr:',mr,'coarse density=',partsum
    call mpi_reduce(partsum,partot,1,MPI_DOUBLE_PRECISION,MPI_SUM,0, &
                    MPI_COMM_WORLD,ierr)
    if (master) write(*,*) 'average coarse grid density=',partot

  end subroutine density_check 
#endif

!*******

  subroutine coarse_power

    implicit none
    include 'mpif.h'
    include 'globalpm.fh'

    integer(4), parameter :: hc=LC/2
    integer(4) :: i,j,k,kg
    real(4) :: kz,ky,kx,kr,k1,k2,w1,w2,pow
    character(len=7) :: z_s
    integer(4) :: fstat,ierr
    character(len=MSL) :: ofile
    real(4) :: rho_c_mean,sum_od

    rho_c_mean=(RLPS/2)**3*MP/LC**3

    call filefftw(LC,rhoc,-1)
    rhoc=rhoc/rho_c_mean-1.0
    call mpi_reduce(sum(rhoc),sum_od,1,mpi_real,mpi_sum,0,mpi_comm_world,ierr)
    if (mr == 0) print *,'sum(overdensity)=',sum_od
    call filefftw(LC,rhoc,1)

    ps_c=0.0

    do k=1,LC/NN
      kg=k+(LC/NN)*mr
      if (kg < hc+2) then
        kz=kg-1
      else
        kz=kg-1-LC
      endif
      do j=1,LC
        if (j < hc+2) then
          ky=j-1
        else
          ky=j-1-LC
        endif
        do i=1,LC+2,2
          kx=(i-1)/2.0
          kr=sqrt(kx**2+ky**2+kz**2)
          if (kr /= 0.0) then
            k1=ceiling(kr)
            k2=k1+1
            w1=k1-kr
            w2=1-w1
            pow=(rhoc(i,j,k)/LC**3)**2+(rhoc(i+1,j,k)/LC**3)**2
            ps_c(1,k1)=ps_c(1,k1)+w1
            ps_c(2,k1)=ps_c(2,k1)+w1*pow
            ps_c(1,k2)=ps_c(1,k2)+w2
            ps_c(2,k2)=ps_c(2,k2)+w2*pow
          endif
        enddo
      enddo
    enddo

    ps_c_sum=0.0
    call mpi_reduce(ps_c,ps_c_sum,LC*2,mpi_real,mpi_sum,0,mpi_comm_world,ierr)

    if (mr == 0) then
      do k=1,LC
        if (ps_c_sum(1,k) /= 0) then
          ps_c_sum(2,k)=4.0*acos(-1.0)*((k-1)**3)*ps_c_sum(2,k)/ps_c_sum(1,k)
          ps_c_sum(1,k)=2.0*acos(-1.0)*(k-1)/box
        endif
      enddo
      
      write(z_s,'(f7.3)') 1/((aGdt/dt)*(6*acos(-1.0))) - 1.0
      z_s=adjustl(z_s)
      ofile=OUT1//z_s(1:len_trim(z_s))//'ps.dat'
      open(50,file=ofile,status='replace',iostat=fstat,form='formatted')

      if (fstat /= 0) then
        write(*,*) 'error opening power spectrum file for write'
        write(*,*) 'rank',mr,'file:',ofile
        call mpi_abort(mpi_comm_world,ierr,ierr)
      endif

      do k=1,LC
        write(50,'(2f20.10)') ps_c_sum(:,k)
      enddo

      close(50)

    endif

  end subroutine coarse_power

!*******

  subroutine coarse_force

    implicit none
    include 'globalpm.fh'

    integer(kind=4) :: i,j,k,ii,ir

    call filefftw(LC,rhoc,1)

    !$omp parallel do default(shared) private(i,j,k,ii,ir)
    do k=1,LC/NN
      do j=1,LC
        do i=1,LC/2+1
          ii=2*i
          ir=ii-1
          force_c(:,ir,j,k)=-rhoc(ii,j,k)*cgk(:,i,j,k)
          force_c(:,ii,j,k)=rhoc(ir,j,k)*cgk(:,i,j,k)
        enddo
      enddo
    enddo
    !$omp end parallel do

    temp_force(:,:,:)=force_c(1,:,:,1:LPSSC)
    call filefftw(LC,temp_force,-1)
    force_c(1,:,:,1:LPSSC)=temp_force(:,:,:)

    temp_force(:,:,:)=force_c(2,:,:,1:LPSSC)
    call filefftw(LC,temp_force,-1)
    force_c(2,:,:,1:LPSSC)=temp_force(:,:,:)

    temp_force(:,:,:)=force_c(3,:,:,1:LPSSC)
    call filefftw(LC,temp_force,-1)
    force_c(3,:,:,1:LPSSC)=temp_force(:,:,:)    

  end subroutine coarse_force

!*******

  subroutine force_buf

    implicit none
    include 'mpif.h'
    include 'globalpm.fh'
   
    integer(kind=4) :: j
    integer(kind=4) :: status(MPI_STATUS_SIZE)
    integer(kind=4) :: tag,ierr,rank_to,rank_from,buf_size

    buf_size=3*LC*LC

    tag=42

    !$omp parallel do default(shared) private(j)
    do j=1,LC
        force_buffer(:,:,j)=force_c(:,1:LC,j,1)
    enddo
    !$omp end parallel do
    
    rank_to=mod(mr+NN-1,NN)
    rank_from=mod(mr+1,NN)

    call mpi_sendrecv_replace(force_buffer,buf_size,MPI_REAL, &
        rank_to,tag,rank_from,tag,MPI_COMM_WORLD,status,ierr)

    force_c(:,1:LC,:,LPSSC+1)=force_buffer(:,:,:)

    tag=54
 
    !$omp parallel do default(shared) private(j) 
    do j=1,LC
        force_buffer(:,:,j)=force_c(:,1:LC,j,LPSSC)
    enddo
    !$omp end parallel do

    rank_to=mod(mr+1,NN)
    rank_from=mod(mr+NN-1,NN) 
 
    call mpi_sendrecv_replace(force_buffer,buf_size,MPI_REAL, &
        rank_to,tag,rank_from,tag,MPI_COMM_WORLD,status,ierr) 

    force_c(:,1:LC,:,0)=force_buffer(:,:,:)

  end subroutine force_buf

!*******

  subroutine max_coarse_dt

    implicit none
    include 'mpif.h'
    include 'globalpm.fh'

    integer(kind=4) :: i,j,k,ierr
    real(kind=4) :: force,max_force

    max_force=0.0

    !$omp parallel do default(shared) &
    !$omp private(i,j,k,force) &
    !$omp reduction(MAX:max_force)
    do k=1,LPSSC
      do j=1,LC
        do i=1,LC
          force=force_c(1,i,j,k)*force_c(1,i,j,k)+ &
                force_c(2,i,j,k)*force_c(2,i,j,k)+ &
                force_c(3,i,j,k)*force_c(3,i,j,k)
          force=sqrt(force)
          if (force.gt.max_force) max_force=force
        enddo
      enddo
    enddo
    !$omp end parallel do

    max_force=sqrt(4.0/(max_force*aGdt/dt))
    call mpi_reduce(max_force,dt_acc_c,1,mpi_real,mpi_max,0, &
                    mpi_comm_world,ierr)
    call mpi_bcast(dt_acc_c,1,mpi_real,0,mpi_comm_world,ierr)

    if (master) write(*,*) 'maximum dt from coarse grid=',dt_acc_c

  end subroutine max_coarse_dt

!*******

  subroutine coarse_velocity

    implicit none
    include 'globalpm.fh'

    integer(kind=4) :: i,j,k,ip,im,jp,jm,kp,km,i1,i2,j2,j1,k1,k2
    integer(kind=4) :: il,jl
    real(kind=4) :: x,y,z,dx1,dx2,dy1,dy2,dz1,dz2,dV
    real(kind=4), parameter :: RLC=LC

    !$omp parallel do default(shared) &
    !$omp private(ip,i,j,k,x,y,z,i1,i2,dx1,dx2,j1,j2,dy1,dy2, &
    !$omp&              k1,k2,dz1,dz2,dV)
    do k=1,LPSSC
      do j=1,LC
        do i=1,LC
          ip=hoc(i,j,k,1)
          do
            ! use goto *gasp* instead of exit because of ifort v8
            !if (ip.eq.0) exit
            if (ip.eq.0) goto 123
            x=mod(0.25*xvp(1,ip)-0.5+RLC,RLC)
            y=mod(0.25*xvp(2,ip)-0.5+RLC,RLC)
            z=(0.25*xvp(3,ip))-0.5

            i1=floor(x)+1
            i2=mod(i1,LC)+1
            dx1=i1-x
            dx2=1-dx1
            j1=floor(y)+1
            j2=mod(j1,LC)+1
            dy1=j1-y
            dy2=1-dy1
            k1=floor(z)+1
            k2=k1+1
            dz1=k1-z
            dz2=1-dz1

            dV=aGdtC*dx1*dy1*dz1
            xvp(4:6,ip)=xvp(4:6,ip)+force_c(:,i1,j1,k1)*dV
            dV=aGdtC*dx2*dy1*dz1
            xvp(4:6,ip)=xvp(4:6,ip)+force_c(:,i2,j1,k1)*dV
            dV=aGdtC*dx1*dy2*dz1
            xvp(4:6,ip)=xvp(4:6,ip)+force_c(:,i1,j2,k1)*dV
            dV=aGdtC*dx2*dy2*dz1
            xvp(4:6,ip)=xvp(4:6,ip)+force_c(:,i2,j2,k1)*dV
            dV=aGdtC*dx1*dy1*dz2
            xvp(4:6,ip)=xvp(4:6,ip)+force_c(:,i1,j1,k2)*dV
            dV=aGdtC*dx2*dy1*dz2
            xvp(4:6,ip)=xvp(4:6,ip)+force_c(:,i2,j1,k2)*dV
            dV=aGdtC*dx1*dy2*dz2
            xvp(4:6,ip)=xvp(4:6,ip)+force_c(:,i1,j2,k2)*dV
            dV=aGdtC*dx2*dy2*dz2
            xvp(4:6,ip)=xvp(4:6,ip)+force_c(:,i2,j2,k2)*dV

            ip=ll(ip)
          enddo
 123      continue
        enddo
      enddo
    enddo
    !$omp end parallel do

  end subroutine coarse_velocity
